import pandas as pd
import torch
from sklearn.model_selection import KFold

from sampo.scheduler.selection.metrics import one_hot_encode
from sampo.scheduler.selection.neural_net import NeuralNetTrainer, NeuralNetType


def cross_val_score(X: pd.DataFrame,
                    y: pd.DataFrame,
                    model: NeuralNetTrainer,
                    epochs: int = 100,
                    folds: int = 5,
                    shuffle: bool = False,
                    random_state: int | None = None,
                    type_task: NeuralNetType = None) \
        -> tuple[list[float] | dict[str, float], int, NeuralNetTrainer]:
    """
    Evaluate metric by cross-validation and also record score times.

    :param X: The data to fit (DataFrame).
    :param y: The column that contains target variable to try to predict.
    :param model: The object (inherited from nn.Module).
    :param epochs: Number of epochs during which the model is trained.
    :param folds: Training dataset is split on 'folds' folds for cross-validation.
    :param shuffle: 'True' if the splitting dataset on folds should be random, 'False' - otherwise.
    :param random_state:
    :return: List of scores that correspond to each validation fold.
    """
    kf = KFold(n_splits=folds, shuffle=shuffle, random_state=random_state)
    scores = 0
    best_loss = 0
    best_trainer: NeuralNetTrainer | None = None
    transform_data = lambda x: one_hot_encode(x, 2) if type_task == NeuralNetType.CLASSIFICATION else x

    for fold, (train_idx, test_idx) in enumerate(kf.split(X)):
        train_tensor = torch.stack([torch.Tensor(v) for v in X.iloc[train_idx, :].values])
        train_target_tensor = torch.stack([torch.Tensor(transform_data(v)) for v in y.iloc[train_idx].values])
        test_tensor = torch.stack([torch.Tensor(v) for v in X.iloc[test_idx, :].values])
        test_target_tensor = torch.stack([torch.Tensor(transform_data(v)) for v in y.iloc[test_idx].values])

        model.fit(train_tensor, train_target_tensor, epochs)
        tmp_score, loss = model.validate(test_tensor, test_target_tensor)
        if tmp_score > scores:
            best_trainer = model
            scores = max(scores, tmp_score)
            best_loss = loss

    return scores, best_loss, best_trainer



from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph
from sampo.schemas.schedule import Schedule
from sampo.schemas.serializable import AutoJSONSerializable
from sampo.utilities.schedule import fix_split_tasks, offset_schedule
from sampo.utilities.serializers import custom_serializer


class ScheduledProject(AutoJSONSerializable['ScheduledProject']):
    def __init__(self, wg: WorkGraph, contractors: list[Contractor], schedule: Schedule):
        self.schedule = schedule
        self.schedule_df = fix_split_tasks(schedule.full_schedule_df)
        self.wg = wg
        self.contractors = contractors

    @custom_serializer('contractors')
    def serialize_contractors(self, value):
        return [v._serialize() for v in value]

    @classmethod
    @custom_serializer('contractors', deserializer=True)
    def deserialize_equipment(cls, value):
        return [Contractor._deserialize(v) for v in value]

    def set_start_date(self, start_date: str) -> 'ScheduledProject':
        self.schedule_df = offset_schedule(self.schedule_df, start_date)
        return self



from sampo.schemas.graph import GraphNode, WorkGraph
from sampo.structurator.prepare_wg_copy import prepare_work_graph_copy, new_start_finish


def graph_in_graph_insertion(master_wg: WorkGraph, master_start: GraphNode, master_finish: GraphNode,
                             slave_wg: WorkGraph, change_id: bool = True) -> WorkGraph:
    """
    Inserts the slave WorkGraph into the master WorkGraph,
    while the starting vertex slave_wg becomes the specified master_start,
    and the finishing vertex is correspondingly master_finish
    :param master_wg: the WorkGraph into which the insertion is performed
    :param master_start: GraphNode which will become the parent for the entire slave_wg
    :param master_finish: GraphNode which will become a child for the whole slave_wg
    :param slave_wg: WorkGraph to be inserted into master_wg
    :param change_id: do ids in the new graph need to be changed
    :return: new union WorkGraph
    """
    master_nodes, master_old_to_new_ids = prepare_work_graph_copy(master_wg, change_id=change_id)
    slave_nodes, slave_old_to_new = prepare_work_graph_copy(slave_wg, [slave_wg.start, slave_wg.finish], change_id)

    # add parent links from slave_wg's nodes to master_start node from slave_wg's nodes
    master_start = master_nodes[master_old_to_new_ids[master_start.id]]
    for edge in slave_wg.start.edges_from:
        slave_nodes[slave_old_to_new[edge.finish.id]].add_parents([master_start])

    # add parent links from master_finish to slave_wg's nodes
    master_finish = master_nodes[master_old_to_new_ids[master_finish.id]]
    master_finish.add_parents([slave_nodes[slave_old_to_new[edge.start.id]]
                               for edge in slave_wg.finish.edges_to])

    start, finish = new_start_finish(master_wg, master_nodes, master_old_to_new_ids)
    return WorkGraph(start, finish)



import random
from abc import abstractmethod, ABC
from dataclasses import dataclass
from random import Random
from typing import Optional

from sampo.schemas.serializable import AutoJSONSerializable

# to work with float and avoid errors due to inaccuracy
EPS = 1e5
# to work with distributions when certain start and finish value is undefined
INF = float('inf')
MINUS_INF = float('-inf')


# TODO: Take out common parts, deal with types, rethink intervals


class Interval(AutoJSONSerializable['BaseReq'], ABC):
    """
    A class for generating random numbers from a given boundary and distribution.
    """

    @abstractmethod
    def rand_float(self, rand: Optional[random.Random] = None) -> float:
        """
        Returns a random float in the interval boundary according to the distribution of the class

        :param rand: object for generating random numbers
        :return kind: the random float
        """
        ...

    @abstractmethod
    def rand_int(self, rand: Optional[random.Random] = None) -> int:
        """
        Returns a random int in the interval boundary according to the distribution of the class

        :param rand: object for generating random numbers
        :return kind: the random int
        """
        ...


@dataclass(frozen=True)
class IntervalUniform(Interval):
    """
    Implementation for uniform distribution

    :param min_val: left border for the interval
    :param max_val: right  border for the interval
    :param rand: object for generating random numbers with, if you want to use a randomizer with a determined seed
    """
    min_val: float
    max_val: float
    rand: Optional[Random] = Random()

    def rand_float(self, rand: Optional[random.Random] = None) -> float:
        """
        Returns a random float in the interval boundary according to the distribution of the class

        :param rand: object for generating random numbers, if you want to use a randomizer with a determined seed
        :return kind: the random float
        """
        rand = rand or self.rand
        return rand.uniform(self.min_val, self.max_val)

    def rand_int(self, rand: Optional[random.Random] = None) -> int:
        """
        Returns a random int in the interval boundary according to the distribution of the class

        :param rand: object for generating random numbers, if you want to use a randomizer with a determined seed
        :return kind: the random int
        """
        rand = rand or self.rand
        value = round(rand.uniform(self.min_val, self.max_val))
        # TODO fix cast math.inf to int
        value = max(value, int(self.min_val - EPS))
        value = min(value, int(self.max_val + EPS))
        return value


@dataclass(frozen=True)
class IntervalGaussian(Interval):
    """
    Implementation for Gaussian distribution

    :param mean: mean for the distribution
    :param sigma: variance for the distribution
    :param min_val: left border for the interval
    :param max_val: right  border for the interval
    :param rand: object for generating random numbers with, if you want to use a randomizer with a determined seed
    """

    mean: float
    sigma: float
    min_val: Optional[float] = MINUS_INF
    max_val: Optional[float] = INF
    rand: Optional[Random] = Random()

    def rand_float(self, rand: Optional[random.Random] = None) -> float:
        """
        Returns a random float in the interval boundary according to the distribution of the class

        :param rand: object for generating random numbers
        :return kind: the random float
        """
        rand = rand or self.rand
        return rand.gauss(self.mean, self.sigma)

    def rand_int(self, rand: Optional[random.Random] = None) -> int:
        """
        Returns a random int in the interval boundary according to the distribution of the class

        :param rand: object for generating random numbers
        :return kind: the random int
        """
        rand = rand or self.rand
        value = round(rand.gauss(self.mean, self.sigma))
        # TODO fix cast math.inf to int
        value = max(value, int(self.min_val - EPS))
        value = min(value, int(self.max_val + EPS))
        return value



import pickle
from random import Random
from typing import Optional

from sampo.generator.environment.contractor import get_contractor, get_contractor_with_equal_proportions
from sampo.generator.pipeline.project import get_graph, SyntheticGraphType
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph


# TODO check outer usage
def graph_from_file(filepath: str, number_of_workers_in_contractors: int) -> tuple[WorkGraph, list[Contractor]]:
    with open(filepath, 'rb') as f:
        data = pickle.load(f)

    wg: WorkGraph = data['work_graph']
    contractors: list[Contractor] = get_contractor_with_equal_proportions(number_of_workers_in_contractors)

    return wg, contractors


# TODO check outer usage
# Functions for generating synthetic graphs
def generate_work_graph(graph_mode: SyntheticGraphType, bottom_border: int,
                        rand: Optional[Random] = None) -> WorkGraph:
    return get_graph(mode=graph_mode,
                     bottom_border=int(bottom_border),
                     addition_cluster_probability=0,
                     rand=rand)


def generate_resources_pool(contractor_capacity: int, num_contractors: int = 1) -> list[Contractor]:
    return [get_contractor(pack_size, index=i) for i, pack_size in enumerate([contractor_capacity] * num_contractors)]



from typing import Type, Callable

from sampo.scheduler.base import SchedulerType
from sampo.scheduler.generic import GenericScheduler
from sampo.scheduler.heft.prioritization import prioritization
from sampo.scheduler.resource.base import ResourceOptimizer
from sampo.scheduler.resource.full_scan import FullScanResourceOptimizer
from sampo.scheduler.timeline.just_in_time_timeline import JustInTimeTimeline
from sampo.scheduler.timeline.momentum_timeline import MomentumTimeline
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator
from sampo.utilities.base_opt import dichotomy_int


class HEFTScheduler(GenericScheduler):
    """
    Scheduler that uses method of a critical path.
    The scheduler gives opportunity to add work only to end.
    """

    def __init__(self,
                 scheduler_type: SchedulerType = SchedulerType.HEFTAddEnd,
                 resource_optimizer: ResourceOptimizer = FullScanResourceOptimizer(),
                 timeline_type: Type = JustInTimeTimeline,
                 work_estimator: WorkTimeEstimator = DefaultWorkEstimator(),
                 prioritization_f: Callable = prioritization,
                 resource_optimize_f: Callable = None):
        if resource_optimize_f is None:
            resource_optimize_f = self.get_default_res_opt_function()
        super().__init__(scheduler_type, resource_optimizer, timeline_type,
                         prioritization_f, resource_optimize_f, work_estimator)
        self._timeline_type = timeline_type


class HEFTBetweenScheduler(HEFTScheduler):
    """
    Type of scheduler that use method of critical path.
    The scheduler give opportunity to add work between existing works.
    """

    def __init__(self,
                 scheduler_type: SchedulerType = SchedulerType.HEFTAddBetween,
                 resource_optimizer: ResourceOptimizer = FullScanResourceOptimizer(),
                 work_estimator: WorkTimeEstimator = DefaultWorkEstimator()):
        super().__init__(scheduler_type, resource_optimizer, MomentumTimeline,
                         resource_optimize_f=self.get_default_res_opt_function(self.get_finish_time),
                         work_estimator=work_estimator)

    @staticmethod
    def get_finish_time(node, worker_team, node2swork, spec, assigned_parent_time, timeline, work_estimator):
        return timeline.find_min_start_time_with_additional(node, worker_team, node2swork, spec, None,
                                                            assigned_parent_time, work_estimator)[1]



import numpy as np
from collections import defaultdict
from sortedcontainers import SortedList
from typing import Iterable

from sampo.schemas.schedule import Schedule
from sampo.schemas.time import Time


def get_total_resources_usage(schedule: Schedule, resources_names: Iterable[str] | None = None) -> dict[str, np.ndarray]:
    df = schedule.full_schedule_df
    points = df[['start', 'finish']].to_numpy().copy()
    points[:, 1] += 1
    points = SortedList(set(points.flatten()))
    usage = defaultdict(lambda: np.zeros_like(points))

    is_none = resources_names is None
    resources_names = set(resources_names) if not is_none else {}

    for swork in schedule.works:
        start = points.bisect_left(swork.start_time)
        finish = points.bisect_left(swork.finish_time + 1)
        for worker in swork.workers:
            if worker.name in resources_names or is_none:
                usage[worker.name][start: finish] += worker.count

    return usage


def get_resources_peak_usage(schedule: Schedule, resources_names: Iterable[str] | None = None) -> dict[str, int]:
    return {res: max(res_usage) for res, res_usage in get_total_resources_usage(schedule, resources_names).items()}


def resources_peaks_sum(schedule: Schedule, resources_names: Iterable[str] | None = None) -> int:
    """
    Count the summary of resources peaks usage in received schedule
    """
    if schedule.execution_time.is_inf():
        return Time.inf().value
    return sum(get_resources_peak_usage(schedule, resources_names).values())


def resources_sum(schedule: Schedule, resources_names: Iterable[str] | None = None) -> int:
    """
    Count the summary usage of resources in received schedule
    """
    is_none = resources_names is None
    resources_names = set(resources_names) if not is_none else {}

    res_sum = sum([sum([worker.count * work.duration.value for worker in work.workers
                        if worker.name in resources_names or is_none], start=0)
                   for work in schedule.works])

    return res_sum


def resources_costs_sum(schedule: Schedule, resources_names: Iterable[str] | None = None) -> float:
    """
    Count the summary cost of resources in received schedule
    """
    is_none = resources_names is None
    resources_names = set(resources_names) if not is_none else {}

    cost = sum([sum([worker.get_cost() * work.duration.value for worker in work.workers
                     if worker.name in resources_names or is_none], start=0.0)
                for work in schedule.works])

    return cost



from dataclasses import dataclass
from typing import TypeVar, Iterable, Generic, Callable

T = TypeVar('T')


@dataclass
class Node(Generic[T]):
    value: T
    next: 'Node' = None


class Iterator(Generic[T]):
    def __init__(self, lst: 'LinkedList'):
        self._lst = lst
        self._node = lst._root
        self._prev = None

    def __iter__(self):
        yield self._node

        while self._node is not None:
            yield self.__next__()

    def __next__(self) -> Node[T]:
        return self.next()

    def next(self) -> Node[T]:
        self._prev = self._node
        self._node = self._node.next
        return self._node

    def get(self) -> Node[T]:
        return self._node

    def has_next(self) -> bool:
        return self._node is not None

    def remove(self) -> T:
        if self._node == self._lst._root:
            old = self._lst._root
            self._lst._root = old.next
            self._node = self._lst._root
            self._lst._len -= 1
            return old.value
        node = self._node
        next = node.next
        self._prev.next = next
        self._node = next
        self._lst._len -= 1
        return node.value


class LinkedList(Generic[T]):
    def __init__(self, iterable: Iterable[T] | None = None):
        self._root = None
        self._tail = None
        self._len = 0
        if iterable:
            for v in iterable:
                self.append(v)

    def append(self, v: T):
        if self._root is None:
            self._root = Node(v)
            self._tail = self._root
        else:
            old_tail = self._tail
            self._tail = Node(v)
            old_tail.next = self._tail
        self._len += 1

    def remove_if(self, condition: Callable[[T], bool]):
        it = self.iterator()

        while it.has_next():
            v = it.get()
            if condition(v.value):
                it.remove()
            else:
                it.next()

    def __add__(self, other: T):
        self.append(other)

    def iterator(self):
        return Iterator(self)

    def __iter__(self):
        return (node.value for node in self.iterator())

    def __len__(self):
        return self._len

    def is_empty(self):
        return len(self) == 0
